Type: Package
Package: actiread
Title: Baseline Package for Reading Actigraphy and Activity Data
Version: 0.3.0
Authors@R: c(
    person(given = "John",
           family = "Muschelli",
           role = c("aut", "cre"),
           email = "muschellij2@gmail.com",
           comment = c(ORCID = "0000-0001-6469-1750"))
    )
Description: Provides baseline functions for reading actigraphy and activity
    data, relying on baseline functions from 'actibase'.  Reads data from 
    'Axivity' 'CWA' <https://axivity.com/> , 
    ActiGraph 'GT3X' <https://ametris.com/actigraph-wgt3x-bt>, 
    'SensorLog' <https://sensorlog.berndthomas.net/>, and 
    'SensorLogger' <https://www.tszheichoi.com/sensorlogger> zipped CSV files.
License: GPL-3
Depends: R (>= 2.10)
Suggests: testthat, utils, covr, knitr, rmarkdown
Encoding: UTF-8
VignetteBuilder: knitr
Imports: actibase, GGIRread, read.gt3x (>= 0.2.0), R.utils, magrittr,
        dplyr, lubridate, assertthat, cli, tibble, janitor, purrr,
        readr, methods
URL: https://jhuwit.github.io/actiread/
BugReports: https://github.com/jhuwit/actiread/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-15 16:02:27 UTC; johnmuschelli
Author: John Muschelli [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6469-1750>)
Maintainer: John Muschelli <muschellij2@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-15 17:30:12 UTC
Built: R 4.6.1; ; 2026-07-15 18:28:34 UTC; unix

# Generated by roxygen2: do not edit by hand

export(acti_convert_sensorlogger_time)
export(acti_example_cwa)
export(acti_example_gt3x)
export(acti_example_sensorlog_file)
export(acti_example_sensorlogger_file)
export(acti_example_sensorlogger_location_file)
export(acti_info_gt3x)
export(acti_read_cwa)
export(acti_read_gt3x)
export(acti_read_sensorlog)
export(acti_read_sensorlogger)
export(acti_read_sensorlogger_general)
export(acti_read_sensorlogger_location)
export(acti_rewrite_sensorlog_csv)
export(acti_sensorlog_csv_colnames_mapping)
export(acti_sensorlog_csv_spec)
export(acti_sensorlogger_location_colnames_mapping)
export(acti_sensorlogger_location_spec)
export(tzoffset_to_tz)
import(actibase)
importFrom(magrittr,"%>%")

<!-- badges: start -->

[![R-CMD-check](https://github.com/jhuwit/actiread/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/jhuwit/actiread/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://app.codecov.io/gh/jhuwit/actiread/branch/main/graph/badge.svg)](https://app.codecov.io/gh/jhuwit/actiread?branch=main)
<!-- badges: end -->

<!-- README.md is generated from README.Rmd. Please edit that file -->

# actiread Package

`actiread` reads raw actigraphy files and returns tidy data frames that
are ready for downstream analysis.

It currently focuses on:

- reading `gt3x` files
- reading `cwa` files
- Reading [SensorLog](https://sensorlog.berndthomas.net/) files
- Reading [SensorLogger](https://www.tszheichoi.com/sensorlogger) files
- loading small example files bundled with the package
- inspecting GT3X header metadata
- handling timezone conversion during import

## Installation

You can install `actiread` from GitHub with:

``` r
# install.packages("remotes")
remotes::install_github("jhuwit/actiread")
```

## Quick start

``` r
library(actibase)
library(actiread)

gt3x_file = acti_example_gt3x()
cwa_file = acti_example_cwa()

gt3x = acti_read_gt3x(gt3x_file, verbose = FALSE)
cwa = acti_read_cwa(cwa_file, verbose = FALSE)

head(gt3x)
#> # A tibble: 6 × 4
#>   time                    X      Y     Z
#>   <dttm>              <dbl>  <dbl> <dbl>
#> 1 2019-09-17 22:40:00 0      0.008 0.996
#> 2 2019-09-17 22:40:00 0.016  0     1.01 
#> 3 2019-09-17 22:40:00 0.02  -0.008 1.00 
#> 4 2019-09-17 22:40:00 0.016 -0.012 1.01 
#> 5 2019-09-17 22:40:00 0.016 -0.008 1.01 
#> 6 2019-09-17 22:40:00 0.008 -0.008 1.01
head(cwa)
#> # A tibble: 6 × 7
#>   time                    x      y      z  temp battery light
#>   <dttm>              <dbl>  <dbl>  <dbl> <dbl>   <dbl> <dbl>
#> 1 2019-02-26 10:55:06 0.328  0.984  0.203  25.6       0   283
#> 2 2019-02-26 10:55:06 0.822 -0.343 -0.368  25.6       0   283
#> 3 2019-02-26 10:55:06 0.874 -0.390 -0.390  25.6       0   283
#> 4 2019-02-26 10:55:06 0.890 -0.391 -0.391  25.6       0   283
#> 5 2019-02-26 10:55:06 0.891 -0.376 -0.391  25.6       0   283
#> 6 2019-02-26 10:55:06 0.876 -0.360 -0.376  25.6       0   283

res = acti_standardize_data(gt3x)
resampled = acti_resample(res, sample_rate = 30L)
get_transformations(resampled)
#> [1] "acti_resample:sample_rate_attribute_changed_to_30"
#> [2] "acti_resample:linear_resampled_to_30Hz"           
#> [3] "acti_read_gt3x:timezone_GMT_forced"               
#> [4] "acti_read_gt3x:timezone_Etc/GMT-4_applied"        
#> [5] "acti_read_gt3x:attributes_set"                    
#> [6] "acti_fill_zeros:filled_zeros"                     
#> [7] "acti_read_gt3x:data_read"
```

``` r
tzoffset_to_tz(c("+00:00", "-05:00", "+01:00"))
#> [1] "Etc/GMT0"  "Etc/GMT-5" "Etc/GMT+1"
```

## What you can do

`actiread` is a focused import layer for actigraphy data:

- read `.gt3x`/`.gt3x.gz` files with `acti_read_gt3x()`
- read `.cwa`/`.cwa.gz` files with `acti_read_cwa()`
- inspect GT3X metadata with `acti_info_gt3x()`
- use bundled sample files with `acti_example_gt3x()` and
  `acti_example_cwa()`
- control timezone handling during import with `tz` and `apply_tz`

The readers return tidy objects with attached metadata such as the file
header and sampling rate, which makes it straightforward to hand the
result off to your own cleaning, summarization, or modeling code.

#' Rewrite a CSV that may have issues with export from SensorLog
#'
#' @param file Input CSV file
#' @param outfile Output CSV file
#' @param verbose Print Diagnostic messages
#'
#' @return A file path to the new CSV
#' @export
#' @examples
#' sl_file = actiread::acti_example_sensorlog_file()
#' tfile = tempfile()
#' files = utils::unzip(sl_file, exdir = tfile)
#' result = actiread::acti_rewrite_sensorlog_csv(files)
acti_rewrite_sensorlog_csv = function(
    file,
    outfile = tempfile(fileext = ".csv"),
    verbose = FALSE
) {
  if (verbose) {
    message(
      paste0("File being processed:\n",
             file)
    )
  }
  # read the files in
  txt = readLines(file)
  cn = strsplit(txt[1], ",")[[1]]
  ind = which(cn ==  "pedometerFloorDescended(N)")
  ind2 = which(cn ==  "pedometerEndDate(txt)")
  # different line enders
  if (length(ind) > 0) {
    stopifnot(ind2 == (ind+1))
  }

  # fix for oceans
  bat_ind = which(cn ==   "batteryTimeStamp_since1970(s)")
  bat_ind_r = which(cn ==  "batteryState(R)")
  bat_ind_z = which(cn ==  "batteryLevel(Z)")
  if (length(bat_ind) > 0) {
    stopifnot(bat_ind_z == (bat_ind+2))
    stopifnot(bat_ind_r == (bat_ind+1))
    stopifnot(bat_ind_z == length(cn) ||
                bat_ind_z == (length(cn) - 1) )
  }

  # split the data in ,
  ss = strsplit(txt, ",")

  # find elements with "null" as they cause issues
  add_empty = sapply(ss, function(x) tolower(x)[ind] == "null")
  l = sapply(ss, length)
  max_length = max(l)
  # stopifnot(max(max_length - l) <= 1)
  ss = lapply(ss, function(x) {
    if (tolower(x)[ind] == "null" &&
        length(x) %in% c(max_length - 1, max_length - 2)) {
      x = c(x[1:ind], "", x[(ind+1):length(x)])
    }
    if (tolower(x)[bat_ind] == "" &&
        length(x) == max_length - 1) {
      x = c(x[1:bat_ind_r], "")
    }
    x
  })
  # make sure consistent with the data
  l = sapply(ss, length)
  stopifnot(all(l == max_length))
  ss = sapply(ss, paste, collapse = ",")
  # write it out
  writeLines(ss, outfile)
  outfile
}

#' Transform timezone offset to timezone
#'
#' @param x A character vector
#'
#' @returns A character vector
#' @export
#'
#' @examples
#' tzoffset_to_tz(c("+00:00", "-05:00", "+01:00"))
tzoffset_to_tz = function(x) {
  stopifnot(all(grepl(":00", x)))
  x = sub(":00:00$", "", x)
  x = sub(":00$", "", x)
  stopifnot(nchar(x) <= 3)
  x = as.numeric(x)
  x = ifelse(x > 0, paste0("+", x), as.character(x))
  x = paste0("Etc/GMT", x)
  stopifnot(x %in% OlsonNames())
  x
}


read_csv_safe = function(..., guess_max = Inf) {
  x = readr::read_csv(..., guess_max = guess_max)
  p = readr::problems(x)
  cn = list(...)$col_names
  if (is.null(cn)) {
    cn = TRUE
  }
  if (nrow(p) > 0) {
    message(paste(utils::capture.output(print(p)), collapse = "\n"))

    rows = unique(p$row)
    if (cn) {
      rows = rows - 1
    }
    bad_data = x[rows, unique(p$col)]
    message("Bad Data:\n")
    message(paste(utils::capture.output(print(bad_data)), collapse = "\n"))
  }
  readr::stop_for_problems(x)
  x
}

test_that("csv helper reports and returns parsed data", {
  valid_file <- tempfile(fileext = ".csv")
  writeLines(c("a,b", "1,2"), valid_file)
  parsed <- read_csv_safe(valid_file, show_col_types = FALSE)
  expect_equal(nrow(parsed), 1L)
  expect_equal(names(parsed), c("a", "b"))

  bad_header_file <- tempfile(fileext = ".csv")
  writeLines(c("a,b", "1,2", "3,bad"), bad_header_file)
  expect_error(
    read_csv_safe(
      bad_header_file,
      col_types = readr::cols(a = readr::col_double(), b = readr::col_double()),
      show_col_types = FALSE
    )
  )

  bad_no_header_file <- tempfile(fileext = ".csv")
  writeLines(c("1,2", "3,bad"), bad_no_header_file)
  expect_error(
    read_csv_safe(
      bad_no_header_file,
      col_names = FALSE,
      col_types = readr::cols(.default = readr::col_double()),
      show_col_types = FALSE
    )
  )
})

acti_process_header = function(data) {
  header = data$header
  if (is.null(header)) {
    header = attr(data, "header")
  }
  if (is.null(header)) {
    header = list()
  }
  if (is.list(data) && !is.null(data$data)) {
    data = data$data
    if (length(header) == 0) {
      header = attr(data, "header")
    }
  }

  if (!is.null(header$accrange)) {
    header$acceleration_min = paste0("-", header$accrange)
    header$acceleration_max = as.character(header$accrange)
  }
  header$sample_rate = header$frequency
  if (is.null(header$sample_rate)) {
    header$sample_rate = attr(data, "sample_rate")
  }
  header
}

# zip_file = "data/SensorLog/OCEANS_011_BL/SensorLogFiles_OCEANS_011_BL_221208_11-55-16.zip"

is_zip_file = function(file) {
  ext = tolower(tools::file_ext(file))
  ext == "zip"
}

unzip_files = function(file) {
  if (any(is_zip_file(file))) {
    if (!all(is_zip_file(file))) {
      stop(paste0("acti_read_sensorlog works with only zip file or a vector of ",
                  "csv files"))
    }
    orig_file = file
    file = lapply(file, function(r) {
      tfile = tempfile()
      utils::unzip(r, exdir = tfile)
    })
    file = unlist(file)
  }
  file
}

#' Read SensorLog Data
#'
#' @param file A character vector of SensorLog files, usually from unzipping
#' the file
#' @param verbose print diagnostic messages.  Either logical or integer, where
#' higher values are higher levels of verbosity.
#' @param robust if `TRUE` then [acti_rewrite_sensorlog_csv] is run on the data
#' to try to fix any shifts with the data.
#' @return A `data.frame` of data
#' @export
#' @examples
#' library(actiread)
#' file = acti_example_sensorlog_file()
#' df = acti_read_sensorlog(file)
#' head(df)
acti_read_sensorlog = function(
    file,
    verbose = FALSE,
    robust = FALSE
) {
  file = unzip_files(file)
  lon_zero = lat_zero = lat = lon = NULL
  rm(list = c("lat", "lon", "lat_zero", "lon_zero"))

  names(file) = file

  cn = acti_sensorlog_csv_colnames_mapping()
  spec = acti_sensorlog_csv_spec()

  if (robust) {
    file = sapply(file, acti_rewrite_sensorlog_csv, verbose = verbose > 1)
  }
  data =
    purrr::map_df(names(file), function(nx) {
      x = file[[nx]]
      r = read_csv_safe(x, progress = FALSE, col_types = spec)
      # r = readr::read_csv(x, col_types = spec)
      if (nrow(r) == 0) {
        return(NULL)
      }
      make_na_cols = c("locationSpeedAccuracy(m/s)")
      missing_cols = setdiff(cn, colnames(r))
      if (length(missing_cols) > 0) {
        missing_cols = unique(missing_cols)
        msg = paste0("Missing expected columns from ", nx, ": ",
                     paste(missing_cols, collapse = ", ")
        )
        message(msg)
        warning(msg)
        for (icol in intersect(missing_cols, make_na_cols)) {
          r[[icol]] = NA_real_
        }
      }
      r = r[, cn]
      colnames(r) = names(cn)

      # r = r %>%
      #   dplyr::mutate(
      #     lat = ifelse(abs(lat) < 0.00001, NA_real_, lat),
      #     lon = ifelse(abs(lon) < 0.00001, NA_real_, lon)
      #   )

      # should add lat_zero in there
      r = r %>%
        dplyr::mutate(
          lat_zero = abs(lat) < 0.00001,
          lon_zero = abs(lon) < 0.00001
        )
      r = r %>%
        dplyr::mutate(
          lat = ifelse(lat_zero, NA_real_, lat),
          lon = ifelse(lon_zero, NA_real_, lon)
        )
      r = r %>%
        dplyr::mutate(
          file = nx
        )

      r
    }, .progress = verbose > 0)
  if (nrow(data) > 0) {
    data = data %>%
      dplyr::select(file, dplyr::everything())
  }
  data
}






#' @export
#' @rdname acti_read_sensorlog
acti_sensorlog_csv_spec = function() {
  spec = readr::cols(
    `loggingTime(txt)` = readr::col_character(),
    `loggingSample(N)` = readr::col_double(),
    `locationTimestamp_since1970(s)` = readr::col_double(),
    `locationLatitude(WGS84)` = readr::col_double(),
    `locationLongitude(WGS84)` = readr::col_double(),
    `locationAltitude(m)` = readr::col_double(),
    `locationSpeed(m/s)` = readr::col_double(),
    `locationSpeedAccuracy(m/s)` = readr::col_double(),
    `accelerometerAccelerationX(G)` = readr::col_double(),
    `accelerometerAccelerationY(G)` = readr::col_double(),
    `accelerometerAccelerationZ(G)` = readr::col_double(),
    .default = readr::col_character()
  )
  spec
}

#' @export
#' @rdname acti_read_sensorlog
acti_sensorlog_csv_colnames_mapping = function() {
  cn =  c(
    time = "loggingTime(txt)",
    index = "loggingSample(N)",
    timestamp = "locationTimestamp_since1970(s)",
    lat = "locationLatitude(WGS84)",
    lon = "locationLongitude(WGS84)",
    altitude = "locationAltitude(m)",
    speed = "locationSpeed(m/s)",
    speed_accuracy = "locationSpeedAccuracy(m/s)",
    accel_X = "accelerometerAccelerationX(G)",
    accel_Y = "accelerometerAccelerationY(G)",
    accel_Z = "accelerometerAccelerationZ(G)")
  cn
}

testthat::context("SensorLog and SensorLogger readers")

write_named_csv <- function(dir, filename, lines) {
  path = file.path(dir, filename)
  writeLines(lines, path)
  path
}

extract_zip_files <- function(zip_file) {
  exdir = tempfile("sensorlogger-")
  dir.create(exdir)
  utils::unzip(zip_file, exdir = exdir)
  list.files(exdir, recursive = TRUE, full.names = TRUE)
}

extdata_file <- function(filename) {
  path = system.file("extdata", filename, package = "actiread")
  if (nzchar(path)) {
    return(path)
  }
  normalizePath(file.path("..", "..", "inst", "extdata", filename), mustWork = TRUE)
}

testthat::test_that("SensorLog helpers and schemas are available", {
  zip_file = actiread::acti_example_sensorlog_file()
  if (!nzchar(zip_file)) {
    zip_file = extdata_file("SensorLogFiles_my_iOS_device_250311_14-55-58.zip")
  }
  sensorlogger_zip = actiread::acti_example_sensorlogger_file()
  if (!nzchar(sensorlogger_zip)) {
    sensorlogger_zip = extdata_file("SensorLogger-2025-04-28_22-04-35.zip")
  }
  sensorlogger_location = actiread::acti_example_sensorlogger_location_file()
  if (!nzchar(sensorlogger_location)) {
    sensorlogger_location = extdata_file("SensorLogger_Location.csv.gz")
  }

  csv_file = tempfile(fileext = ".csv")
  writeLines("a,b\n1,2", csv_file)

  testthat::expect_true(file.exists(zip_file))
  testthat::expect_true(file.exists(sensorlogger_zip))
  testthat::expect_true(file.exists(sensorlogger_location))
  testthat::expect_true(actiread:::is_zip_file(zip_file))
  testthat::expect_false(actiread:::is_zip_file(csv_file))
  testthat::expect_identical(actiread:::unzip_files(csv_file), csv_file)

  testthat::expect_error(
    actiread:::unzip_files(c(zip_file, csv_file)),
    "only zip file or a vector of csv files"
  )

  sensorlog_cn = actiread:::acti_sensorlog_csv_colnames_mapping()
  sensorlog_spec = actiread:::acti_sensorlog_csv_spec()
  sensorlogger_cn = actiread:::acti_sensorlogger_location_colnames_mapping()
  sensorlogger_spec = actiread:::acti_sensorlogger_location_spec()

  testthat::expect_identical(
    names(sensorlog_cn),
    c(
      "time", "index", "timestamp", "lat", "lon", "altitude",
      "speed", "speed_accuracy", "accel_X", "accel_Y", "accel_Z"
    )
  )
  testthat::expect_true("loggingTime(txt)" %in% names(sensorlog_spec$cols))
  testthat::expect_identical(
    names(sensorlogger_cn),
    c(
      "time", "seconds_elapsed", "altitude", "speed_accuracy",
      "bearing_accuracy", "lat", "altitude_above_mean_sea_level", "bearing",
      "horizontal_accuracy", "vertical_accuracy", "lon", "speed"
    )
  )
  testthat::expect_true("latitude" %in% names(sensorlogger_spec$cols))

  testthat::expect_equal(
    format(actiread::acti_convert_sensorlogger_time(1e9), tz = "UTC"),
    "1970-01-01 00:00:01"
  )
})

testthat::test_that("SensorLog rewrite fixes pedometer and battery shifts", {
  pedometer_dir = tempfile("sensorlog-pedometer-")
  dir.create(pedometer_dir)
  pedometer_file = write_named_csv(
    pedometer_dir,
    "pedometer.csv",
    c(
      "id,pedometerFloorDescended(N),pedometerEndDate(txt),extra",
      "1,null,2025-03-11"
    )
  )
  pedometer_out = actiread:::acti_rewrite_sensorlog_csv(
    pedometer_file,
    verbose = TRUE
  )
  testthat::expect_equal(
    readLines(pedometer_out),
    c(
      "id,pedometerFloorDescended(N),pedometerEndDate(txt),extra",
      "1,null,,2025-03-11"
    )
  )

  battery_dir = tempfile("sensorlog-battery-")
  dir.create(battery_dir)
  battery_file = write_named_csv(
    battery_dir,
    "battery.csv",
    c(
      "id,pedometerFloorDescended(N),pedometerEndDate(txt),batteryTimeStamp_since1970(s),batteryState(R),batteryLevel(Z)",
      "1,0,2025-03-11,,1"
    )
  )
  battery_out = actiread:::acti_rewrite_sensorlog_csv(battery_file)
  testthat::expect_equal(
    readLines(battery_out),
    c(
      "id,pedometerFloorDescended(N),pedometerEndDate(txt),batteryTimeStamp_since1970(s),batteryState(R),batteryLevel(Z)",
      "1,0,2025-03-11,,1,"
    )
  )
})

testthat::test_that("SensorLog reader handles the shipped archive, missing columns, and empty files", {
  example_zip = extdata_file("SensorLogFiles_my_iOS_device_250311_14-55-58.zip")
  example_data = actiread::acti_read_sensorlog(example_zip)

  testthat::expect_s3_class(example_data, "tbl_df")
  testthat::expect_true(nrow(example_data) > 0)
  testthat::expect_true(all(c("file", "time", "lat", "lon", "lat_zero", "lon_zero") %in% names(example_data)))
  testthat::expect_false(anyNA(example_data$lat_zero))
  testthat::expect_false(anyNA(example_data$lon_zero))
  testthat::expect_equal(length(unique(example_data$file)), 1L)
  testthat::expect_match(unique(example_data$file), "my_iOS_device.csv$")

  example_files = actiread:::unzip_files(example_zip)
  testthat::expect_true(length(example_files) == 1L)

  missing_dir = tempfile("sensorlog-missing-")
  dir.create(missing_dir)
  missing_file = write_named_csv(
    missing_dir,
    "SensorLog.csv",
    c(
      paste(
        c(
          "loggingTime(txt)",
          "loggingSample(N)",
          "locationTimestamp_since1970(s)",
          "locationLatitude(WGS84)",
          "locationLongitude(WGS84)",
          "locationAltitude(m)",
          "locationSpeed(m/s)",
          "accelerometerAccelerationX(G)",
          "accelerometerAccelerationY(G)",
          "accelerometerAccelerationZ(G)",
          "pedometerFloorDescended(N)",
          "pedometerEndDate(txt)"
        ),
        collapse = ","
      ),
      "2025-03-11 14:44:11,1,1741707851,0,0,1.2,-1,0.41,0.12,0.34,null,2025-03-11"
    )
  )

  testthat::expect_warning(
    missing_data <- actiread::acti_read_sensorlog(missing_file, robust = TRUE),
    "Missing expected columns"
  )
  testthat::expect_equal(nrow(missing_data), 1L)
  testthat::expect_true(is.na(missing_data$speed_accuracy[1]))
  testthat::expect_true(is.na(missing_data$lat[1]))
  testthat::expect_true(is.na(missing_data$lon[1]))
  testthat::expect_true(missing_data$lat_zero[1])
  testthat::expect_true(missing_data$lon_zero[1])
  testthat::expect_equal(missing_data$file[1], missing_file)

  empty_file = write_named_csv(
    missing_dir,
    "empty.csv",
    c(
      "loggingTime(txt),loggingSample(N),locationTimestamp_since1970(s),locationLatitude(WGS84),locationLongitude(WGS84),locationAltitude(m),locationSpeed(m/s),locationSpeedAccuracy(m/s),accelerometerAccelerationX(G),accelerometerAccelerationY(G),accelerometerAccelerationZ(G)"
    )
  )
  empty_data = actiread::acti_read_sensorlog(empty_file)
  testthat::expect_equal(nrow(empty_data), 0L)
  testthat::expect_equal(ncol(empty_data), 0L)
})

testthat::test_that("SensorLogger helpers normalize stubs and convert time", {
  testthat::expect_equal(
    actiread:::acti_sensorlogger_stub("SensorLogger_Accelerometer.csv"),
    "accelerometer"
  )
  testthat::expect_equal(
    actiread:::acti_sensorlogger_stub("SensorLogger_AccelerometerUncalibrated.csv"),
    "accelerometer_uncalibrated"
  )
  testthat::expect_equal(
    actiread:::acti_sensorlogger_stub("SensorLogger_Location.csv.gz"),
    "location"
  )
  testthat::expect_equal(
    format(actiread::acti_convert_sensorlogger_time(1e9), tz = "UTC"),
    "1970-01-01 00:00:01"
  )
})

testthat::test_that("SensorLogger readers work for the shipped files and dispatch correctly", {
  location_file = extdata_file("SensorLogger_Location.csv.gz")
  location_data = actiread::acti_read_sensorlogger_location(
    location_file,
    show_col_types = FALSE
  )

  testthat::expect_s3_class(location_data, "tbl_df")
  testthat::expect_true(inherits(location_data$time, "POSIXct"))
  testthat::expect_true(all(c("file", "cat_type_sensor", "lat_zero", "lon_zero") %in% names(location_data)))
  testthat::expect_true(all(location_data$file == location_file))
  testthat::expect_equal(unique(location_data$cat_type_sensor), "location")

  general_dir = tempfile("sensorlogger-general-")
  dir.create(general_dir)
  general_file = write_named_csv(
    general_dir,
    "Accelerometer.csv",
    c(
      "time,x,y",
      "1000000000,1,2"
    )
  )
  general_data = actiread::acti_read_sensorlogger_general(
    general_file,
    show_col_types = FALSE
  )
  testthat::expect_s3_class(general_data, "tbl_df")
  testthat::expect_true(inherits(general_data$time, "POSIXct"))
  testthat::expect_equal(unique(general_data$cat_type_sensor), "accelerometer")
  testthat::expect_equal(general_data$file[1], general_file)

  no_time_file = write_named_csv(
    general_dir,
    "Other.csv",
    c(
      "foo,bar",
      "1,2"
    )
  )
  no_time_data = actiread:::acti_sensorlogger_reader(
    no_time_file,
    type = "unknown"
  )
  testthat::expect_s3_class(no_time_data, "tbl_df")
  testthat::expect_false("time" %in% names(no_time_data))
  testthat::expect_equal(unique(no_time_data$cat_type_sensor), "other")

  single_file_data = actiread::acti_read_sensorlogger(
    general_file,
    show_col_types = FALSE
  )
  testthat::expect_s3_class(single_file_data, "tbl_df")
  testthat::expect_equal(single_file_data$file[1], general_file)
  testthat::expect_true(inherits(single_file_data$time, "POSIXct"))

  archive_data = actiread::acti_read_sensorlogger(
    extdata_file("SensorLogger-2025-04-28_22-04-35.zip"),
    show_col_types = FALSE
  )
  testthat::expect_type(archive_data, "list")
  testthat::expect_equal(length(archive_data), 16L)
  testthat::expect_true(all(c("location", "accelerometer", "battery", "gravity") %in% names(archive_data)))
  testthat::expect_true(inherits(archive_data$location$time, "POSIXct"))
})

testthat::test_that("SensorLogger location reader returns NULL for empty files", {
  empty_dir = tempfile("sensorlogger-empty-")
  dir.create(empty_dir)
  empty_file = write_named_csv(
    empty_dir,
    "Location.csv",
    c(
      "time,seconds_elapsed,altitude,speedAccuracy,bearingAccuracy,latitude,altitudeAboveMeanSeaLevel,bearing,horizontalAccuracy,verticalAccuracy,longitude,speed"
    )
  )

  testthat::expect_null(
    actiread::acti_read_sensorlogger_location(empty_file, show_col_types = FALSE)
  )
})

#' Read SensorLogger Data
#'
#' @param file A character vector of SensorLogger files, usually from unzipping
#' the file, or a zip file of SensorLogger files
#' @param verbose print diagnostic messages.  Either logical or integer, where
#' higher values are higher levels of verbosity.
#' @param ... additional arguments to pass to [readr::read_csv()].
#' If `verbose = FALSE`, then `progress = FALSE` and `show_col_types = FALSE`,
#' unless otherwise overridden
#' @return A `data.frame` of data
#' @export
acti_read_sensorlogger = function(
    file,
    verbose = FALSE,
    ...
) {
  lon_zero = lat_zero = lat = lon = NULL
  rm(list = c("lat", "lon", "lat_zero", "lon_zero"))

  file = unzip_files(file)
  stub = acti_sensorlogger_stub(file)
  names(file) = stub

  data_list = purrr::map(file, function(r) {
    data = acti_sensorlogger_reader(r, verbose = verbose, ...,  type = NULL)
  })
  if (length(file) == 1 && length(data_list) == 1) {
    data_list = data_list[[1]]
  }

  data_list
}


#' @inherit actibase::as_datetime_safe
#' @export
acti_convert_sensorlogger_time = function(x) {
  as_datetime_safe(x/1000/1000/1000)
}



acti_sensorlogger_stub = function(x) {
  stub = sub("[.]csv($|[.]gz$)", "", basename(x), ignore.case = TRUE)
  stub = tolower(stub)
  stub = sub("^sensorlogger_", "", stub)
  stub = sub("uncalibrated", "_uncalibrated", stub)
  stub
}




#' @export
#' @rdname acti_read_sensorlogger
acti_read_sensorlogger_general = function(file, ..., verbose = FALSE) {
  df = read_csv_safe(file, ...)
  if (nrow(df) == 0) {
    return(NULL)
  }
  df = df %>%
    janitor::clean_names()
  if (assertthat::has_name(df, "time")) {
    df$time = acti_convert_sensorlogger_time(df$time)
  }

  df$file = file
  stub = acti_sensorlogger_stub(file)
  df$cat_type_sensor = stub

  if (nrow(df) > 0) {
    df = df %>%
      dplyr::select(file, dplyr::everything())
  }
  df
}

acti_sensorlogger_reader = function(file, ..., type = NULL, verbose = FALSE) {
  if (is.null(type)) {
    type = acti_sensorlogger_stub(file)
  }
  func = switch(
    type,
    accelerometer = acti_read_sensorlogger_general,
    accelerometer_uncalibrated = acti_read_sensorlogger_general,
    annotation = acti_read_sensorlogger_general,
    battery = acti_read_sensorlogger_general,
    gravity = acti_read_sensorlogger_general,
    gyroscope = acti_read_sensorlogger_general,
    gyroscope_uncalibrated = acti_read_sensorlogger_general,
    location = acti_read_sensorlogger_location,
    metadata = acti_read_sensorlogger_general,
    orientation = acti_read_sensorlogger_general,
    pedometer = acti_read_sensorlogger_general,
    acti_read_sensorlogger_general
    )
  args = list(...)
  if (!verbose & !"progress" %in% names(args)) {
    args$progress = FALSE
  }
  if (!verbose & !"show_col_types" %in% names(args)) {
    args$show_col_types = FALSE
  }
  args$file = file
  do.call(func, args = args)
}

.read_cwa = function(
    path,
    start = 0,
    end = Inf,
    tz = "UTC",
    ...
) {

  for (ext in c("bz2", "gz", "xz")) {
    if (
      R.utils::isCompressedFile(
        path,
        method = "extension",
        ext = ext,
        fileClass = "")
    ) {
      FUN = switch(ext,
                   gz = gzfile,
                   xz = xzfile,
                   bz2 = bzfile
      )
      path = R.utils::decompressFile(
        path,
        destname = tempfile(fileext = ".cwa"),
        temporary = TRUE,
        overwrite = TRUE,
        ext = ext,
        FUN = FUN,
        remove = FALSE)
      on.exit(unlink(path, recursive = TRUE),
              add = TRUE)
      break
    }
  }
  data = GGIRread::readAxivity(
    filename = path,
    start = start,
    end = end,
    desiredtz = tz,
    ...
  )
  data
}


#' Read CWA File
#'
#' @param path Path to cwa file
#' @param start where to start in the file, passed to [GGIRread::readAxivity]
#' @param end where to end in the file, passed to [GGIRread::readAxivity]
#' @param tz time zone for the data `time`, passed to `desiredtz` argument in
#' [GGIRread::readAxivity].  If NULL or `""`, no time conversion is done.
#' @param apply_tz turn the `time` column into a `POSIXct` and apply the
#' timezone
#' @param ... additional arguments to pass to [GGIRread::readAxivity()]
#' @param verbose print diagnostic messages, higher values = more verbosity.
#' @returns A `tibble` with attributes of a header, sample rate, and
#' transformations
#' @export
#'
#' @examples
#' data = acti_read_cwa(acti_example_cwa())
#' data = acti_read_cwa(
#'   acti_example_cwa(),
#'   tz = NULL,
#'   apply_tz = FALSE,
#'   verbose = FALSE
#' )
acti_read_cwa = function(
    path,
    start = 0,
    end = Inf,
    tz = "UTC",
    ...,
    apply_tz = TRUE,
    verbose = TRUE
) {
  tz_read = tz
  if (is.null(tz_read)) {
    tz_read = ""
  }
  args = list(
    path,
    start = start,
    end = end,
    tz = tz_read,
    ...
  )
  if (verbose && !"progressBar" %in% names(args)) {
    args$progressBar = TRUE
  }
  data = do.call(.read_cwa, args = args)
  acti_cwa_process_time(
    data = data,
    tz = tz,
    apply_tz = apply_tz,
    verbose = verbose
  )
}


.acti_cwa_count_formatted_na = function(
    time,
    tz,
    chunk_size = 1e5
) {
  if (length(time) == 0) {
    return(0L)
  }

  na_count = 0L
  starts = seq.int(1L, length(time), by = chunk_size)
  for (start in starts) {
    end = min(start + chunk_size - 1L, length(time))
    na_count = na_count + sum(is.na(format(time[start:end], tz = tz)))
  }

  na_count
}


# Internal helper used by acti_read_cwa()
acti_cwa_process_time = function(
    data,
    tz = "UTC",
    apply_tz = TRUE,
    verbose = TRUE
) {
  if (is.null(tz)) {
    tz = ""
  }
  assertthat::assert_that(
    assertthat::is.string(tz)
  )

  header = acti_process_header(data)
  if (is.list(data) && !is.null(data$data)) {
    data = data$data
  }
  if (is.null(header)) {
    header = acti_process_header(data)
  }
  data = data %>%
    dplyr::as_tibble()
  attr(data, "header") = header
  attr(data, "sample_rate") = header$sample_rate
  data = set_transformations(data,
                             "acti_read_cwa:data_read_via_readAxivity",
                             add = TRUE)

  time_na_count = sum(is.na(data$time))
  if (apply_tz) {
    if (tz != "") {
      data$time = as.POSIXct(data$time, origin = "1970-01-01",
                             tz = tz)
      data = set_transformations(data,
                                 "acti_read_cwa:converted_timestamp_to_time",
                                 add = TRUE)
    }
    if (verbose) {
      cli::cli_alert_info("Timezone applied to data")
    }
    formatted_na_count = .acti_cwa_count_formatted_na(data$time, tz = tz)
    if (formatted_na_count > time_na_count) {
      stop("Applying timezone from offset created NA times - stopping.")
    }
  } else if (verbose) {
    cli::cli_alert_info("Timezone not applied to data")
  }

  time1 = data$time[1]
  header_start = header$start
  if (is.null(header_start)) {
    header_start = header$start_time
  }
  if (!is.null(header_start) && length(header_start) > 0 && header_start != time1) {
    msg = paste0("Header start date is not same time as data$time,",
                 " may want to use apply_tz = FALSE.")
    warning(msg)
  }

  any_na_time = anyNA(data$time)
  if (any_na_time) {
    warning("Some missing times in cwa data - please check.")
  }

  data = tibble::as_tibble(data)
  attr(data, "header") = header
  attr(data, "sample_rate") = header$sample_rate
  data
}

#' Activity/Actigraphy Example Data
#'
#' @returns A file path
#' @export
#' @rdname acti_example_data
#'
#' @examples
#' library(actiread)
#' acti_example_gt3x()
#' acti_example_cwa()
#' acti_example_sensorlogger_file()
acti_example_gt3x = function() {
  system.file("extdata",
              "TAS1H30182785_2019-09-17.gt3x.gz",
              package = "actiread")
}

#' @rdname acti_example_data
#' @export
acti_example_cwa = function() {
  system.file("extdata",
              "ax3_testfile.cwa.gz",
              package = "actiread")
}


#' @rdname acti_example_data
#' @export
acti_example_sensorlog_file = function() {
  base::system.file(
    "extdata", "SensorLogFiles_my_iOS_device_250311_14-55-58.zip",
    package = "actiread")
}

#' @rdname acti_example_data
#' @export
acti_example_sensorlogger_file = function() {
  base::system.file(
    "extdata", "SensorLogger-2025-04-28_22-04-35.zip",
    package = "actiread")
}

#' @rdname acti_example_data
#' @export
acti_example_sensorlogger_location_file = function() {
  base::system.file(
    "extdata", "SensorLogger_Location.csv.gz",
    package = "actiread")
}

#' Read GT3X file
#'
#' @param path Path to gt3x file
#' @param asDataFrame convert to an `activity_df`, see
#' \code{as.data.frame.activity}
#' @param imputeZeroes Impute zeros in case there are missingness?
#' Default is `FALSE`, in which case
#' the time series will be incomplete in case there is missingness.
#' @param ... additional arguments to pass to [read.gt3x::read.gt3x()]
#' @param verbose print diagnostic messages, higher values = more verbosity.
#' @param apply_tz Apply the timezone from the header `TimeZone` attribute,
#' if available
#' @param check_attributes Check that the attributes are included This is a sanity check,
#' including checking that `sample_rate` is in the attributes.
#' @param tz timezone to project the data into.  The data read in via
#' [read.gt3x::read.gt3x()] says the timezone is GMT, but the time values is in the
#' native timezone.  So this data is projected into the correct time zone and then
#' forced into the timezone given by `tz`.  Set to `NULL` to not apply this
#' forcing.
#' @param fill_zeroes Rows with all zeros will be filled in with the last
#' observation carried forward as is done with ActiLife.  Recommended
#' @returns A `data.frame`
#' @export
#'
#' @examples
#' library(actiread)
#' data = acti_read_gt3x(acti_example_gt3x())
#' data = acti_read_gt3x(
#'   acti_example_gt3x(),
#'   tz = NULL,
#'   apply_tz = FALSE,
#'   verbose = FALSE,
#'   fill_zeroes = FALSE
#' )
acti_read_gt3x = function(
    path,
    asDataFrame = TRUE,
    imputeZeroes = TRUE,
    verbose = TRUE,
    ...,
    fill_zeroes = TRUE,
    apply_tz = FALSE,
    check_attributes = TRUE,
    tz = "GMT"
) {
  args = list(path = path,
              asDataFrame = asDataFrame,
              imputeZeroes = imputeZeroes,
              verbose = verbose > 1,
              ...)
  if (is.null(args$digits) &&
      (
        utils::packageVersion("read.gt3x") >= package_version("1.3.0")  ||
        "digits" %in% methods::formalArgs(read.gt3x::read.gt3x)
      )
  ) {
    args$digits = 5L
  }
  data = do.call(read.gt3x::read.gt3x, args = args)
  data = set_transformations(data,
                             "acti_read_gt3x:data_read",
                             add = TRUE)
  if (fill_zeroes) {
    if (verbose) {
      cli::cli_alert_info("Filling zeros in data")
    }
    data = acti_fill_zeros(data)
    if (verbose) {
      cli::cli_alert_success("Filled zeros in data")
    }
  }
  acti_gt3x_process_time(
    data = data,
    tz = tz,
    apply_tz = apply_tz,
    check_attributes = check_attributes,
    verbose = verbose
  )
}


# Internal helper used by acti_read_gt3x()
acti_gt3x_process_time = function(
    data,
    tz = "GMT",
    apply_tz = TRUE,
    check_attributes = TRUE,
    verbose = TRUE
) {
  if (is.null(tz)) {
    tz = ""
  }
  assertthat::assert_that(
    assertthat::is.string(tz)
  )

  header = acti_process_header(data)
  if (is.list(data) && !is.null(data$data)) {
    data = data$data
  }
  if (is.null(header)) {
    header = acti_process_header(data)
  }
  data = dplyr::as_tibble(data)

  attr(data, "header") = header
  attr(data, "sample_rate") = header$sample_rate
  data = set_transformations(data,
                             "acti_read_gt3x:attributes_set",
                             add = TRUE)

  any_na_time = anyNA(data$time)
  if (any_na_time) {
    warning("Some missing times in gt3x data - please check.")
  }
  if (apply_tz) {
    if (verbose) {
      cli::cli_alert_info("Timezone applied to data")
    }
    if (NROW(header$TimeZone) == 0 || is.null(header$TimeZone)) {
      cli::cli_warn("No TimeZone found in header data in gt3x.")
    } else {
      tz_from_offset = tzoffset_to_tz(header$TimeZone)
      if (verbose) {
        cli::cli_alert_info("Timezone from header: {header$TimeZone}")
        cli::cli_alert_info("Timezone from offset: {tz_from_offset}")
      }
      data$time = lubridate::with_tz(data$time, tz_from_offset)
      data = set_transformations(data,
                                 paste0("acti_read_gt3x:timezone_", tz_from_offset, "_applied"),
                                 add = TRUE)
      if (!is.null(tz)) {
        data$time = lubridate::force_tz(data$time, tz = tz)
        data = set_transformations(data,
                                   paste0("acti_read_gt3x:timezone_", tz, "_forced"),
                                   add = TRUE)
      }
      if (!any_na_time && anyNA(data$time)) {
        stop("Applying timezone from offset created NA times - stopping.")
      }
    }
  } else if (verbose) {
    cli::cli_alert_info("Timezone not applied to data")
  }

  time1 = data$time[1]
  header_start = header$start
  if (is.null(header_start)) {
    header_start = header$start_time
  }
  if (!is.null(header_start) && length(header_start) > 0 && header_start != time1) {
    msg = paste0("Header start date is not same time as data$time,",
                 " may want to use apply_tz = FALSE.")
    warning(msg)
  }

  any_na_time = anyNA(data$time)
  if (any_na_time) {
    warning("Some missing times in gt3x data - please check.")
  }

  data = as.data.frame(data)
  attr(data, "header") = header
  attr(data, "sample_rate") = header$sample_rate
  if (check_attributes) {
    stopifnot(!is.null(attr(data, "sample_rate")))
  }
  tibble::as_tibble(data)
}



#' @export
#' @rdname acti_read_gt3x
#' @examples
#' info = acti_info_gt3x(acti_example_gt3x())
acti_info_gt3x = function(
    path,
    ...
) {

  data = read.gt3x::parse_gt3x_info(
    path = path,
    ...)
  data
}

testthat::context("Reading baseline data")

testthat::test_that("example data helpers return package files", {
  gt3x = actiread::acti_example_gt3x()
  cwa = actiread::acti_example_cwa()

  testthat::expect_true(file.exists(gt3x))
  testthat::expect_true(file.exists(cwa))
  testthat::expect_match(gt3x, "\\.gt3x\\.gz$")
  testthat::expect_match(cwa, "\\.cwa\\.gz$")
})

testthat::test_that("tzoffset_to_tz handles supported offsets", {
  testthat::expect_equal(actiread:::tzoffset_to_tz("+00:00"), "Etc/GMT0")
  testthat::expect_equal(actiread:::tzoffset_to_tz("-05:00"), "Etc/GMT-5")
  testthat::expect_equal(actiread:::tzoffset_to_tz("+01:00"), "Etc/GMT+1")
  testthat::expect_error(actiread:::tzoffset_to_tz("+05:30"))
})

testthat::test_that("GT3X reading works on the shipped example", {
  args = list(
    actiread::acti_example_gt3x(),
    verbose = FALSE,
    fill_zeroes = FALSE,
    digits = 3L
  )
  if (utils::packageVersion("read.gt3x") < package_version("1.3.0")) {
    args$digits = NULL
  }
  res = do.call(actiread::acti_read_gt3x, args = args)

  testthat::expect_true(all(c("time", "X", "Y", "Z") %in% names(res)))
  testthat::expect_equal(mean(res$X), -0.0742151351351352)
  testthat::expect_equal(attr(res, "sample_rate"), 100L)
  testthat::expect_true("transformations" %in% names(attributes(res)))
})


testthat::test_that("GT3X reading works with digits=5L", {
  testthat::skip_if_not_installed("read.gt3x", minimum_version = "1.3.0")
  res = actiread::acti_read_gt3x(
    actiread::acti_example_gt3x(),
    verbose = FALSE,
    fill_zeroes = FALSE,
    digits = 5L
  )

  testthat::expect_true(all(c("time", "X", "Y", "Z") %in% names(res)))
  testthat::expect_equal(mean(res$X), -0.074214755925156)
  testthat::expect_equal(attr(res, "sample_rate"), 100L)
  testthat::expect_true("transformations" %in% names(attributes(res)))
})

testthat::test_that("GT3X reading fills zeroes on the shipped example", {
  res = suppressMessages(
    actiread::acti_read_gt3x(
      actiread::acti_example_gt3x(),
      verbose = TRUE,
      fill_zeroes = TRUE
    )
  )

  testthat::expect_true(all(c("time", "X", "Y", "Z") %in% names(res)))
  testthat::expect_equal(attr(res, "sample_rate"), 100L)
})

testthat::test_that("GT3X reading handles tz=NULL and apply_tz = FALSE", {
  res = suppressMessages(
    actiread::acti_read_gt3x(
      actiread::acti_example_gt3x(),
      tz = NULL,
      apply_tz = FALSE,
      verbose = TRUE,
      fill_zeroes = FALSE,
      check_attributes = FALSE
    )
  )

  testthat::expect_true(all(c("time", "X", "Y", "Z") %in% names(res)))
})

testthat::test_that("GT3X helper handles an empty header and missing times", {
  fake = list(
    header = NULL,
    data = data.frame(
      time = c(
        as.POSIXct("2019-09-17 18:40:00", tz = "UTC"),
        NA
      ),
      X = c(1, 2),
      Y = c(3, 4),
      Z = c(5, 6)
    )
  )

  res = suppressWarnings(
    actiread:::acti_gt3x_process_time(
      fake,
      apply_tz = FALSE,
      verbose = FALSE,
      check_attributes = FALSE
    )
  )
  testthat::expect_true(all(c("time", "X", "Y", "Z") %in% names(res)))
  testthat::expect_true(anyNA(res$time))
})

testthat::test_that("GT3X helper warns when the header timezone is missing", {
  fake = list(
    header = list(
      TimeZone = NULL,
      accrange = 8,
      frequency = 100L,
      start = as.POSIXct("2019-09-17 18:40:00", tz = "UTC")
    ),
    data = data.frame(
      time = c(
        as.POSIXct("2019-09-17 18:40:00", tz = "UTC"),
        as.POSIXct("2019-09-17 18:40:01", tz = "UTC")
      ),
      X = c(1, 2),
      Y = c(3, 4),
      Z = c(5, 6)
    )
  )

  res = suppressWarnings(
    actiread:::acti_gt3x_process_time(
      fake,
      apply_tz = TRUE,
      verbose = FALSE,
      check_attributes = FALSE
    )
  )
  testthat::expect_true(all(c("time", "X", "Y", "Z") %in% names(res)))
})

testthat::test_that("GT3X helper stops if timezone conversion creates NA times", {
  fake = list(
    header = list(
      TimeZone = "+00:00",
      accrange = 8,
      frequency = 100L,
      start = structure(1e20, class = c("POSIXct", "POSIXt"), tzone = "UTC")
    ),
    data = data.frame(
      time = structure(1e20, class = c("POSIXct", "POSIXt"), tzone = "UTC"),
      X = 1,
      Y = 2,
      Z = 3
    )
  )

  testthat::expect_error(
    actiread:::acti_gt3x_process_time(
      fake,
      apply_tz = TRUE,
      verbose = TRUE,
      check_attributes = FALSE
    ),
    "created NA times"
  )
})

testthat::test_that("GT3X helper reports start mismatches", {
  fake = list(
    header = list(
      TimeZone = NULL,
      accrange = 8,
      frequency = 100L,
      start = as.POSIXct("2019-09-17 18:40:01", tz = "UTC")
    ),
    data = data.frame(
      time = c(
        as.POSIXct("2019-09-17 18:40:00", tz = "UTC"),
        as.POSIXct("2019-09-17 18:40:01", tz = "UTC")
      ),
      X = c(1, 2),
      Y = c(3, 4),
      Z = c(5, 6)
    )
  )

  res = suppressWarnings(
    actiread:::acti_gt3x_process_time(
      fake,
      apply_tz = FALSE,
      verbose = TRUE,
      check_attributes = FALSE
    )
  )
  testthat::expect_true(all(c("time", "X", "Y", "Z") %in% names(res)))
})

testthat::test_that("GT3X metadata can be parsed directly", {
  info = actiread::acti_info_gt3x(actiread::acti_example_gt3x())
  testthat::expect_true(is.list(info))
  testthat::expect_true(length(info) > 0)
})

testthat::test_that("CWA reading works on the shipped example", {
  res = actiread::acti_read_cwa(
    actiread::acti_example_cwa(),
    verbose = FALSE
  )

  testthat::expect_true(all(c("time", "x", "y", "z") %in% names(res)))
  testthat::expect_true(mean(res$x) > -10)
  testthat::expect_equal(attr(res, "sample_rate"), 100L)
  testthat::expect_true("transformations" %in% names(attributes(res)))
})

testthat::test_that("CWA reading handles tz=NULL and apply_tz = FALSE", {
  res = suppressMessages(
    actiread::acti_read_cwa(
      actiread::acti_example_cwa(),
      tz = NULL,
      apply_tz = FALSE,
      verbose = TRUE
    )
  )

  testthat::expect_true(all(c("time", "x", "y", "z") %in% names(res)))
})

testthat::test_that("CWA helper falls back to the header attribute", {
  fake = list(
    data = data.frame(
      time = c(
        as.POSIXct("2020-01-01 00:00:00", tz = "UTC"),
        as.POSIXct("2020-01-01 00:00:01", tz = "UTC")
      ),
      x = c(1, 2),
      y = c(3, 4),
      z = c(5, 6)
    )
  )
  attr(fake, "header") = list(
    accrange = 8,
    frequency = 100L,
    start = as.POSIXct("2020-01-01 00:00:00", tz = "UTC")
  )

  res = suppressWarnings(
    actiread:::acti_cwa_process_time(
      fake,
      apply_tz = FALSE,
      verbose = FALSE
    )
  )

  testthat::expect_true(all(c("time", "x", "y", "z") %in% names(res)))
  testthat::expect_equal(attr(res, "sample_rate"), 100L)
})

testthat::test_that("CWA helper handles tz = NULL", {
  fake = list(
    header = list(
      accrange = 8,
      frequency = 100L,
      start = as.POSIXct("2020-01-01 00:00:00", tz = "UTC")
    ),
    data = data.frame(
      time = c(
        as.POSIXct("2020-01-01 00:00:00", tz = "UTC"),
        as.POSIXct("2020-01-01 00:00:01", tz = "UTC")
      ),
      x = c(1, 2),
      y = c(3, 4),
      z = c(5, 6)
    )
  )

  res = suppressWarnings(
    actiread:::acti_cwa_process_time(
      fake,
      tz = NULL,
      apply_tz = FALSE,
      verbose = FALSE
    )
  )

  testthat::expect_true(all(c("time", "x", "y", "z") %in% names(res)))
})

testthat::test_that("CWA helper handles an empty header", {
  fake = list(
    header = NULL,
    data = data.frame(
      time = c(
        as.POSIXct("2020-01-01 00:00:00", tz = "UTC"),
        as.POSIXct("2020-01-01 00:00:01", tz = "UTC")
      ),
      x = c(1, 2),
      y = c(3, 4),
      z = c(5, 6)
    )
  )

  res = suppressWarnings(
    actiread:::acti_cwa_process_time(
      fake,
      apply_tz = FALSE,
      verbose = FALSE
    )
  )

  testthat::expect_true(all(c("time", "x", "y", "z") %in% names(res)))
})

testthat::test_that("CWA helper reports the no-timezone branch", {
  fake = list(
    header = list(
      accrange = 8,
      frequency = 100L,
      start = as.POSIXct("2020-01-01 00:00:00", tz = "UTC")
    ),
    data = data.frame(
      time = c(
        as.POSIXct("2020-01-01 00:00:00", tz = "UTC"),
        as.POSIXct("2020-01-01 00:00:01", tz = "UTC")
      ),
      x = c(1, 2),
      y = c(3, 4),
      z = c(5, 6)
    )
  )

  res = suppressMessages(
    actiread:::acti_cwa_process_time(
      fake,
      apply_tz = FALSE,
      verbose = TRUE
    )
  )

  testthat::expect_true(all(c("time", "x", "y", "z") %in% names(res)))
  testthat::expect_equal(attr(res, "sample_rate"), 100L)
})

testthat::test_that("CWA helper warns on start mismatch and missing times", {
  fake = list(
    header = list(
      accrange = 8,
      frequency = 100L,
      start = as.POSIXct("2020-01-01 00:00:00", tz = "UTC")
    ),
    data = data.frame(
      time = c(
        as.POSIXct("2020-01-01 01:00:00", tz = "UTC"),
        NA
      ),
      x = c(1, 2),
      y = c(3, 4),
      z = c(5, 6)
    )
  )

  res = suppressWarnings(
    actiread:::acti_cwa_process_time(
      fake,
      apply_tz = FALSE,
      verbose = FALSE
    )
  )

  testthat::expect_true(anyNA(res$time))
  testthat::expect_equal(attr(res, "sample_rate"), 100L)
})

testthat::test_that("CWA helper stops if timezone conversion creates NA times", {
  fake = list(
    header = list(
      accrange = 8,
      frequency = 100L,
      start = structure(1e20, class = c("POSIXct", "POSIXt"), tzone = "UTC")
    ),
    data = data.frame(
      time = c(1e20, 1e20 + 1),
      x = c(1, 2),
      y = c(3, 4),
      z = c(5, 6)
    )
  )

  testthat::expect_error(
    actiread:::acti_cwa_process_time(
      fake,
      apply_tz = TRUE,
      verbose = TRUE
    ),
    "created NA times"
  )
})

testthat::test_that("CWA helper stops when timezone conversion adds missing times to pre-existing NA data", {
  fake = list(
    header = list(
      accrange = 8,
      frequency = 100L,
      start = structure(1e20, class = c("POSIXct", "POSIXt"), tzone = "UTC")
    ),
    data = data.frame(
      time = c(NA_real_, 1e20),
      x = c(1, 2),
      y = c(3, 4),
      z = c(5, 6)
    )
  )

  testthat::expect_error(
    actiread:::acti_cwa_process_time(
      fake,
      apply_tz = TRUE,
      verbose = TRUE
    ),
    "created NA times"
  )
})

testthat::test_that("CWA helper reports timezone application", {
  fake = list(
    header = list(
      accrange = 8,
      frequency = 100L,
      start = as.POSIXct("1970-01-01 00:00:00", tz = "UTC")
    ),
    data = data.frame(
      time = c(0, 1),
      x = c(1, 2),
      y = c(3, 4),
      z = c(5, 6)
    )
  )

  res = suppressMessages(
    actiread:::acti_cwa_process_time(
      fake,
      apply_tz = TRUE,
      verbose = TRUE
    )
  )

  testthat::expect_true(all(c("time", "x", "y", "z") %in% names(res)))
})

#' @export
#' @rdname acti_read_sensorlogger
acti_sensorlogger_location_colnames_mapping = function() {
  # cn =  c(
  #   time = "time",
  #   seconds_elapsed = "seconds_elapsed",
  #   altitude = "altitude",
  #   speedAccuracy = "speed_accuracy",
  #   bearingAccuracy = "bearing_accuracy",
  #   latitude = "lat",
  #   altitudeAboveMeanSeaLevel = "altitude_above_mean_sea_level",
  #   bearing = "bearing",
  #   horizontalAccuracy = "horizontal_accuracy",
  #   verticalAccuracy = "vertical_accuracy",
  #   longitude = "lon",
  #   speed = "speed"
  # )
  cn = c(
    time = "time",
    seconds_elapsed = "seconds_elapsed",
    altitude = "altitude",
    speed_accuracy = "speedAccuracy",
    bearing_accuracy = "bearingAccuracy",
    lat = "latitude",
    altitude_above_mean_sea_level = "altitudeAboveMeanSeaLevel",
    bearing = "bearing",
    horizontal_accuracy = "horizontalAccuracy",
    vertical_accuracy = "verticalAccuracy",
    lon = "longitude",
    speed = "speed"
  )
  cn
}

#' @export
#' @rdname acti_read_sensorlogger
acti_sensorlogger_location_spec = function() {
  spec = readr::cols(
    time = readr::col_double(),
    seconds_elapsed = readr::col_double(),
    altitude = readr::col_double(),
    speedAccuracy = readr::col_double(),
    bearingAccuracy = readr::col_double(),
    latitude = readr::col_double(),
    altitudeAboveMeanSeaLevel = readr::col_double(),
    bearing = readr::col_double(),
    horizontalAccuracy = readr::col_double(),
    verticalAccuracy = readr::col_double(),
    longitude = readr::col_double(),
    speed = readr::col_double()
  )
  spec
}


create_lat_lon_zero = function(df) {
  lon_zero = lat_zero = lat = lon = NULL
  rm(list = c("lat", "lon", "lat_zero", "lon_zero"))

  # should add lat_zero in there
  df = df %>%
    dplyr::mutate(
      lat_zero = abs(lat) < 0.00001,
      lon_zero = abs(lon) < 0.00001
    )
  df = df %>%
    dplyr::mutate(
      lat = ifelse(lat_zero, NA_real_, lat),
      lon = ifelse(lon_zero, NA_real_, lon)
    )
  df
}

#' @export
#' @rdname acti_read_sensorlogger
acti_read_sensorlogger_location = function(file, ...) {
  df = read_csv_safe(file, ...)
  if (nrow(df) == 0) {
    return(NULL)
  }
  cn = acti_sensorlogger_location_colnames_mapping()
  spec = acti_sensorlogger_location_spec()

  icn = intersect(cn, colnames(df))
  df = df[, cn]
  colnames(df) = names(cn)
  df$time = acti_convert_sensorlogger_time(df$time)
  df = create_lat_lon_zero(df)
  df$file = file
  stub = acti_sensorlogger_stub(file)
  df$cat_type_sensor = stub
  df
}

---
title: "Reading and Manipulating Activity Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Reading and Manipulating Activity Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: refs.bib
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = ""
)
```

```{r setup}
library(actibase)
library(actiread)
```

# Overview

`actibase` is the foundation layer for actigraphy and activity data.
It focuses on raw file reading, standardization, calibration, resampling,
activity counts, non-wear detection, and transformation bookkeeping.

Higher-level summarization, step-count mapping, and downstream statistical
analysis are meant to live in overlay packages that build on top of this
core package.

# Reading data

The package ships with a small GT3X file and a small CWA file under
`inst/extdata`. We will use those examples here.

```{r}
gt3x_file = acti_example_gt3x()
cwa_file = acti_example_cwa()

gt3x = acti_read_gt3x(gt3x_file, verbose = FALSE)
cwa = acti_read_cwa(cwa_file, verbose = FALSE)

class(gt3x)
names(gt3x)
class(cwa)
names(cwa)
```

The GT3X reader uses `read.gt3x::read.gt3x()` underneath, while the CWA
reader uses `GGIRread::readAxivity()`.

If you need to inspect the GT3X metadata separately, `acti_info_gt3x()`
parses the header information without returning the full data stream.

```{r}
info = acti_info_gt3x(gt3x_file)
names(info)
```

The readers also let you control timezone handling explicitly. Setting
`apply_tz = FALSE` keeps the timestamps as stored in the file, and `tz = NULL`
disables the final timezone forcing step.

```{r}
gt3x_no_tz = acti_read_gt3x(
  gt3x_file,
  tz = NULL,
  apply_tz = FALSE,
  verbose = FALSE,
  fill_zeroes = FALSE
)

cwa_no_tz = acti_read_cwa(
  cwa_file,
  tz = NULL,
  apply_tz = FALSE,
  verbose = FALSE
)

get_transformations(gt3x_no_tz)
get_transformations(cwa_no_tz)
```

Internally, Axivity files use fixed UTC offsets. The helper that maps those
offsets to Olson timezone names is small but useful when you need to reason
about the conversion logic:

```{r}
tzoffset_to_tz(c("+00:00", "-05:00", "+01:00"))
```

# Standardizing 

The baseline package keeps the data in a consistent shape:

```{r}
std = acti_standardize_data(gt3x)

head(std)
```


# Resampling

You can resample a three-axis signal to a new sampling rate or to specific
timestamps:

```{r}
resampled = acti_resample(std, sample_rate = 30L)
get_transformations(resampled)

same_times = acti_resample_to_time(
  std,
  times = lubridate::floor_date(std$time, unit = "1 second")
)
get_transformations(same_times)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acti_read_sensorlog.R
\name{acti_read_sensorlog}
\alias{acti_read_sensorlog}
\alias{acti_sensorlog_csv_spec}
\alias{acti_sensorlog_csv_colnames_mapping}
\title{Read SensorLog Data}
\usage{
acti_read_sensorlog(file, verbose = FALSE, robust = FALSE)

acti_sensorlog_csv_spec()

acti_sensorlog_csv_colnames_mapping()
}
\arguments{
\item{file}{A character vector of SensorLog files, usually from unzipping
the file}

\item{verbose}{print diagnostic messages.  Either logical or integer, where
higher values are higher levels of verbosity.}

\item{robust}{if \code{TRUE} then \link{acti_rewrite_sensorlog_csv} is run on the data
to try to fix any shifts with the data.}
}
\value{
A \code{data.frame} of data
}
\description{
Read SensorLog Data
}
\examples{
library(actiread)
file = acti_example_sensorlog_file()
df = acti_read_sensorlog(file)
head(df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acti_read_sensorlogger_zzz.R
\name{acti_convert_sensorlogger_time}
\alias{acti_convert_sensorlogger_time}
\title{Convert vectors ensuring no new NA}
\usage{
acti_convert_sensorlogger_time(x)
}
\arguments{
\item{x}{a vector}
}
\value{
A converted `vector` the same length as `x` or errors if there
are introduced NAs.
}
\description{
Convert vectors ensuring no new NA
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acti_read_gt3x.R
\name{acti_read_gt3x}
\alias{acti_read_gt3x}
\alias{acti_info_gt3x}
\title{Read GT3X file}
\usage{
acti_read_gt3x(
  path,
  asDataFrame = TRUE,
  imputeZeroes = TRUE,
  verbose = TRUE,
  ...,
  fill_zeroes = TRUE,
  apply_tz = FALSE,
  check_attributes = TRUE,
  tz = "GMT"
)

acti_info_gt3x(path, ...)
}
\arguments{
\item{path}{Path to gt3x file}

\item{asDataFrame}{convert to an \code{activity_df}, see
\code{as.data.frame.activity}}

\item{imputeZeroes}{Impute zeros in case there are missingness?
Default is \code{FALSE}, in which case
the time series will be incomplete in case there is missingness.}

\item{verbose}{print diagnostic messages, higher values = more verbosity.}

\item{...}{additional arguments to pass to \code{\link[read.gt3x:read.gt3x]{read.gt3x::read.gt3x()}}}

\item{fill_zeroes}{Rows with all zeros will be filled in with the last
observation carried forward as is done with ActiLife.  Recommended}

\item{apply_tz}{Apply the timezone from the header \code{TimeZone} attribute,
if available}

\item{check_attributes}{Check that the attributes are included This is a sanity check,
including checking that \code{sample_rate} is in the attributes.}

\item{tz}{timezone to project the data into.  The data read in via
\code{\link[read.gt3x:read.gt3x]{read.gt3x::read.gt3x()}} says the timezone is GMT, but the time values is in the
native timezone.  So this data is projected into the correct time zone and then
forced into the timezone given by \code{tz}.  Set to \code{NULL} to not apply this
forcing.}
}
\value{
A \code{data.frame}
}
\description{
Read GT3X file
}
\examples{
library(actiread)
data = acti_read_gt3x(acti_example_gt3x())
data = acti_read_gt3x(
  acti_example_gt3x(),
  tz = NULL,
  apply_tz = FALSE,
  verbose = FALSE,
  fill_zeroes = FALSE
)
info = acti_info_gt3x(acti_example_gt3x())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acti_read_sensorlogger_location.R,
%   R/acti_read_sensorlogger_zzz.R
\name{acti_sensorlogger_location_colnames_mapping}
\alias{acti_sensorlogger_location_colnames_mapping}
\alias{acti_sensorlogger_location_spec}
\alias{acti_read_sensorlogger_location}
\alias{acti_read_sensorlogger}
\alias{acti_read_sensorlogger_general}
\title{Read SensorLogger Data}
\usage{
acti_sensorlogger_location_colnames_mapping()

acti_sensorlogger_location_spec()

acti_read_sensorlogger_location(file, ...)

acti_read_sensorlogger(file, verbose = FALSE, ...)

acti_read_sensorlogger_general(file, ..., verbose = FALSE)
}
\arguments{
\item{file}{A character vector of SensorLogger files, usually from unzipping
the file, or a zip file of SensorLogger files}

\item{...}{additional arguments to pass to \code{\link[readr:read_csv]{readr::read_csv()}}.
If \code{verbose = FALSE}, then \code{progress = FALSE} and \code{show_col_types = FALSE},
unless otherwise overridden}

\item{verbose}{print diagnostic messages.  Either logical or integer, where
higher values are higher levels of verbosity.}
}
\value{
A \code{data.frame} of data
}
\description{
Read SensorLogger Data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_utils.R
\name{tzoffset_to_tz}
\alias{tzoffset_to_tz}
\title{Transform timezone offset to timezone}
\usage{
tzoffset_to_tz(x)
}
\arguments{
\item{x}{A character vector}
}
\value{
A character vector
}
\description{
Transform timezone offset to timezone
}
\examples{
tzoffset_to_tz(c("+00:00", "-05:00", "+01:00"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rewrite_csv.R
\name{acti_rewrite_sensorlog_csv}
\alias{acti_rewrite_sensorlog_csv}
\title{Rewrite a CSV that may have issues with export from SensorLog}
\usage{
acti_rewrite_sensorlog_csv(
  file,
  outfile = tempfile(fileext = ".csv"),
  verbose = FALSE
)
}
\arguments{
\item{file}{Input CSV file}

\item{outfile}{Output CSV file}

\item{verbose}{Print Diagnostic messages}
}
\value{
A file path to the new CSV
}
\description{
Rewrite a CSV that may have issues with export from SensorLog
}
\examples{
sl_file = actiread::acti_example_sensorlog_file()
tfile = tempfile()
files = utils::unzip(sl_file, exdir = tfile)
result = actiread::acti_rewrite_sensorlog_csv(files)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acti_example_data.R
\name{acti_example_gt3x}
\alias{acti_example_gt3x}
\alias{acti_example_cwa}
\alias{acti_example_sensorlog_file}
\alias{acti_example_sensorlogger_file}
\alias{acti_example_sensorlogger_location_file}
\title{Activity/Actigraphy Example Data}
\usage{
acti_example_gt3x()

acti_example_cwa()

acti_example_sensorlog_file()

acti_example_sensorlogger_file()

acti_example_sensorlogger_location_file()
}
\value{
A file path
}
\description{
Activity/Actigraphy Example Data
}
\examples{
library(actiread)
acti_example_gt3x()
acti_example_cwa()
acti_example_sensorlogger_file()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/acti_read_cwa.R
\name{acti_read_cwa}
\alias{acti_read_cwa}
\title{Read CWA File}
\usage{
acti_read_cwa(
  path,
  start = 0,
  end = Inf,
  tz = "UTC",
  ...,
  apply_tz = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{path}{Path to cwa file}

\item{start}{where to start in the file, passed to \link[GGIRread:readAxivity]{GGIRread::readAxivity}}

\item{end}{where to end in the file, passed to \link[GGIRread:readAxivity]{GGIRread::readAxivity}}

\item{tz}{time zone for the data \code{time}, passed to \code{desiredtz} argument in
\link[GGIRread:readAxivity]{GGIRread::readAxivity}.  If NULL or \code{""}, no time conversion is done.}

\item{...}{additional arguments to pass to \code{\link[GGIRread:readAxivity]{GGIRread::readAxivity()}}}

\item{apply_tz}{turn the \code{time} column into a \code{POSIXct} and apply the
timezone}

\item{verbose}{print diagnostic messages, higher values = more verbosity.}
}
\value{
A \code{tibble} with attributes of a header, sample rate, and
transformations
}
\description{
Read CWA File
}
\examples{
data = acti_read_cwa(acti_example_cwa())
data = acti_read_cwa(
  acti_example_cwa(),
  tz = NULL,
  apply_tz = FALSE,
  verbose = FALSE
)
}
